#####################################################################
#:TriggerBase.pm
#
# Licensed Materials - Property of IBM
# Rational ClearCase
# (C) Copyright IBM Corp. 2001, 2011. All Rights Reserved.
# US Government Users Restricted Rights
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp
#

=head1 NAME

CQCC::TriggerBase - Class to provide generic services for triggers,
including initialization, processing arguments and other services.

=head1 CLASS DESCRIPTION

This class manages initialization for the integration by reading in
the command line arguments; creating ClearCase and ClearQuest class
instances through IOClassFactory(); and cleaning up and terminating
the trigger session.

A subclass of this trigger is expected to provide the specific
behavior for an application, such as TriggerCQCC.

SUPPORT POLICY: See TriggerCQCC.pm for the Rational Support Policy. 

=head1 METHODS

=cut

######################################################################
#::Package CQCC::TriggerBase
######################################################################

package CQCC::TriggerBase;

use strict;

# Note: These classes are all the base classes that the trigger
# needs directly. At runtime, subclasses may actually be loaded
# and used instead. For example, CQuest is specified here but
# in practice either CQWeb or CQPerlAPI would be used. 

use CQCC::CQuest;
use CQCC::CCase;
use CQCC::IOBase;
use CQCC::Parm;
use CQCC::Timer;

sub Version { return '2.1.1' };

######################################################################
#::Class Globals 
######################################################################

######################################################################
#::function ClassInit

=over 6

=item C<ClassInit()>

This class function is automatically invoked to define class 
parameters using CQCC::Parm objects.

=over 4

=item C<TriggerInterface>

Name of Trigger subclass to run for this session, used by
IOClassFactory(). Set to 'CQCC::Trigger' by default.

=item C<REQUIRED>

Internal flag indicating if the operation is required (TRUE)
or not (FALSE). Set through a command line argument (-required)
but can also be set as a configuration parameter.

=item C<SILENT>

Internal flag to other classes that the session is in silent mode
(TRUE) or not (FALSE). Silent is used for when checkout only interaction 
is selected but the checkin trigger still needs to run without user 
interaction (silently) to cleanup association data. Set through a 
command line argument (-silent).

=item C<TESTMODE>

Internal flag to other classes that the session is in debug mode
(TRUE) or not (FALSE). Set through a command line argument (-test).

=item C<SAVE_SERIES_CACHE>

Internal flag set to indicate if a series cache should be saved.

=item C<HIDE_ERROR_MSG_SPECIFIC>

Internal flag set to indicate if hide some specific error messages
during some process. Currently it used to blind specific error messages
during retry PostCommand in CQWebJava. 

=back

=cut


######################################################################

sub ClassInit { 
    # Environment Parameters

    if (&ParmDefined("TriggerInterface")) {
        return;
    } 

    # Configuration Parameters

    &SetParmDefaults(PCLASS=>"CQCC::TriggerBase", SRC_CFG=>1);
    &DefineParm(NAME=>"TriggerInterface", 
         DEFAULT=>"CQCC::TriggerCQCC", SRC_CMD=>1);

    # Command line options

    &SetParmDefaults(PCLASS=>"CQCC::TriggerBase", SRC_CMD=>1);

    &DefineParm(NAME=>"REQUIRED", SRC_CFG=>1, DEFAULT=>"FALSE",
                RANGE_LIST => ["FALSE", "TRUE"]);
    &DefineParm(NAME=>"SILENT", DEFAULT=>"FALSE",
                RANGE_LIST => ["FALSE", "TRUE"]);
    &DefineParm(NAME=>"TESTMODE", DEFAULT=>"FALSE",
                RANGE_LIST => ["FALSE", "TRUE"]);

    &SetParmDefaults(PCLASS=>"CQCC::TriggerBase", SRC_INT=>1);
    &DefineParm(NAME=>"SAVE_SERIES_CACHE", DEFAULT=>"FALSE",
                RANGE_LIST => ["FALSE", "TRUE"]);

    &DefineParm(NAME=>"HIDE_ERROR_MSG_SPECIFIC", DEFAULT=>0, SRC_INT=>1);

    &SetParmDefaults();
}

&ClassInit();

######################################################################
#::method Connect

=over 6

=item C<Connect(system)>

This method connects to ClearCase or ClearQuest, handling any
necessary initialization of the class interface and storing the 
value into the CC or CQ instance variables respectively. 

=back

=cut 

######################################################################

sub Connect {
    my ($self,$system) = @_;

    if (uc $system eq "CLEARCASE") {
       if (! defined $self->{CC}) {
           $self->{CC} = &IOClassFactory("ClearCaseInterface");
           # Save series cache if in series and not at end
           # Pre-checkin trigger can also turn this on 
           my $series = &GetParm("CLEARCASE_SERIES_ID");
           my $endSeries = &GetParm("CLEARCASE_END_SERIES");
           if ((defined $series && $series ne "" && $series ne "0") &&
               (defined $endSeries && $endSeries ne "1")) {
               &ForceParm("SAVE_SERIES_CACHE","TRUE");
           }
       }
    } elsif (uc $system eq "CLEARQUEST") {
       if (! defined $self->{CQ}) {
           $self->{CQ} = &IOClassFactory("ClearQuestInterface");
           $self->{CQ}->Connect();
       }
    } else {
       &IOMessage("FATAL", ["TBASE_E_UNKNOWN_SYSTEM", $system]);
    }

    return;
}

######################################################################
#::method DispatchOperation

=over 6

=item C<DispatchOperation()>

This method dispatchs the selected operation with its arguments
and exits with the return status of the operation. 

=back

=cut 

######################################################################

sub DispatchOperation {
    my $self = shift;

    my $op = $self->{OPERATION};
    my @args = split(",",$self->{ARGSLIST});
    my @quotedArgs;
    for (@args) {
        push @quotedArgs, "\'$_\'";
    } 
    my $argstring = join(",",@quotedArgs);

    my $cmd = "\$self->Op_$op($argstring)";
    &IODebugPrint("BASIC","TriggerBase::DispatchOperation $cmd\n");

    my $result = eval "$cmd";

    # Disable to hide specific error messages
    &ForceParm("HIDE_ERROR_MSG_SPECIFIC", 0);

    # Operations not returning a distinct result value probably failed
    if (! defined $result ) {
        &IOMessage("ERROR", ["TBASE_E_BAD_RESULT", $result, 
                   $self->{OPERATION}]);
        $result = 1;
    }
    &IOExitHandler($result);
}

######################################################################
#::method GUIDisplay
######################################################################

sub GUIDisplay {
    my ($self) = @_;
    $self->GUIBuild();
    &IOstd()->RunMainLoop();
    return 0;
}

######################################################################
#::method Initialize

=over 6

=item C<Initialize()>

This method will initialize the trigger after it has been created. It
determines which classes should be instantiated, processes command
line arguments, retrieves any cached information since the last
invocation and begins testmode operations if that has been selected.

=back

=cut 

######################################################################

sub Initialize {
    my $self = shift;

    $self->ConfigureClasses();
    $self->ProcessArguments();
    $self->LoadCache();

    &CQCC::CQuest::ValidateDatabaseEntity();
    # Testmode preamble - print parameters, environment info, etc. 
    if (&GetParm("TESTMODE") eq "TRUE" || 
          &GetParm("CQCC_DEBUG") gt 1) {
        my @results = &PrintParms(); # Parm.pm can't ref IOBase.pm
        if (&GetParm("CQCC_DEBUG") gt 1) {
            &IODebugPrint("BASIC",join("\n",@results));
            my @evs = ();
            for (sort keys %ENV) {
                push @evs, "$_: $ENV{$_}";
            }
            &IODebugPrint("VERY_DETAILED",join("\n",@evs));
        } else {
            &IOPrint(join("\n",@results));
        }
    }
}

######################################################################
#::method LoadCache

=over 6

=item C<LoadCache()>

This method reads back the parameter values stored in the current
cache file. On UNIX the cache file (cqcc_cache) is in the home 
directory of the user; on Windows is is kept in the profiles folder.

=back

=cut 

######################################################################

sub LoadCache {
    my $self = shift;
    my @parm_strings;
    my $seriesCache = $self->{CACHE_SERIES_FILE};
    my $series = $ENV{"CLEARCASE_SERIES_ID"};

    ########################################
    # ALWAYS read/delete the ccase_cache for current PID
    # Then if no seriesCache exists for post-op checkin trigger
    #         AND series id is NOT SET
    #         AND same XPN is used 

    my $ccPid = $ENV{"CLEARCASE_PPID"};
    if (defined $ccPid && $ccPid ne "") {
        my $cc_cache = &IOPathname("PARM",".ccase_cache.$ccPid");
        if (-r $cc_cache) {
            my ($altName,$fileXPN) = &IOFileRead($cc_cache,1);
            my $series = $ENV{"CLEARCASE_SERIES_ID"};
            if (! -r $seriesCache &&
                (! defined $series || $series eq "") &&
                $fileXPN eq $ENV{"CLEARCASE_XPN"} &&
                ($self->{POSTOP} eq 1 && 
                 $ENV{"CLEARCASE_OP_KIND"} eq "checkin")) {
                &IODebugPrint("DETAILED","LoadCache\n" .
                    "using ccase_cache: $altName used for $fileXPN");
                $seriesCache = $altName;
            }
        }
    }

    if (-r $seriesCache) {
        # Read and delete the series file if it exists
        @parm_strings = &IOFileRead( $seriesCache,1 );
        &IODebugPrint("DETAILED", "LoadCache reading $seriesCache\n");
    } elsif ((! defined $series || $series eq "") &&
             (&GetParm("CQCC_POSTCHECKIN_COMMIT") eq "TRUE" ||
              &GetParm("CQCC_POSTCHECKIN_VERIFY") eq "TRUE") &&
             ($self->{POSTOP} eq 1 && $ENV{"CLEARCASE_OP_KIND"} eq "checkin")) {
        &IOMessage("FATAL", [ "TBASE_E_LOAD_CACHE_SERIES", $seriesCache]);
    } else {
        my $sessionCache = $self->{CACHE_FILE};
        if (-r $sessionCache) {
            @parm_strings = &IOFileRead( $sessionCache );
            # Screen out all but user preference parameters 
            # to ignore older caches that have series-specific info
            my @userPrefs = &GetParmNameList("SRC_USER_PREF");
            my @userParms;
            for (@parm_strings) {
                my $entry = $_;
                if ($entry =~ /(\S*)=/) {
                    my $key = $1;
                    for (@userPrefs) {
                        if ($_ eq $key) {
                           push @userParms, $entry;
                           last;
                        }
                    }
                }
            }
            @parm_strings = @userParms;
            &IODebugPrint("DETAILED", "LoadCache reading $sessionCache\n");
        } else {
            &IODebugPrint("DETAILED", "LoadCache found no cache to read\n");
            return;
        }
    }    
    for (@parm_strings) { 
        &ParmParse($_);
    }
}

######################################################################
#::method MenuDisplay

=over 6

=item C<MenuDisplay(menuName, {[ initialChoice ]} )>

This method displays the requested menu and handles interaction with
the user. It looks up the menu definition as an instance variable if
the Trigger object, i.e. "MAIN_MENU" refers to $obj->{MAIN_MENU}.

A menu is a list consisting of a title message tag and one or more
choice message tags which can be translated through the message
catalog along with a set of arguments.  For example, [Action_Help,
"main_menu"] might be displayed as "3) Help for main_menu".  If the
title or choice message tags are NOT message tag arrays then the
literal value is shown instead of being looked up through the message
tag.

When the user makes a selection, the menu gets back the message tag.
The first tag value is expected to be a trigger method and that will
be called with the remaining tag values as arguments, i.e.
$obj->Action_Help("main_menu"). This approach allows a subclass to
redefine the menu by editing its definition or to override any of its
activities by overriding the method that gets called back.

The optional initialChoice argument is used to automatically
make an initial selection, calling the Dialog_SelectByQuery for
example or attempting to automatically commit changes when ApplyToAll
was called to process the remaining elements of a batch for example. 
Like other menu selection items, it expects a message tag. 

=back

=cut 

######################################################################

sub MenuDisplay {
    my ($self, $menuName, $initialChoice) = @_;
    my $menuPtr = $self->{$menuName};
    my ($abort, $choice);

    if (! defined $menuPtr) {
        &IOMessage("FATAL", ["TBASE_E_UNKNOWN_MENU", $menuName]);
    }
    my @menu = @{ $menuPtr };

    # Loops forever, expecting we will exit during cancel or Ok action

    while (1) { 
        if (defined $initialChoice ) {
            $choice = $initialChoice;
            $initialChoice = undef;
        } else {
            $self->{HELP_CONTEXT} = $menuName;
            ($abort, $choice) = &IOListSelect(0, @menu);
        } 
        if ($abort) { 
            &IOExitHandler(1);
        } 
        if ($choice) {
            my @choiceMsgTag = @{ $choice };
            my $method = shift @choiceMsgTag;
            # CCPerl/PC doesn't handle Can() method
            # but it is more robust for CQPerl to be able to use it
            if ($] lt "5.006" || $self->can($method)) { 
                my @args;
                for (@choiceMsgTag) { 
                    push @args,"\"$_\"";
                }
                my $cmd = "\$self->$method(" .  join(",",@args) . ")";
                my $result = eval $cmd;
            } 
            else {
                &IOMessage("ERROR", ["TBASE_E_UNKNOWN_METHOD", 
                     $method]);
            }
        }
    }
    return;
}

######################################################################
#::method new

=over 6

=item C<new CQCC::TriggerBase()>

This method presets the objects attributes to read in the basic
command line arguments. The object is not active until the owner
calls it's Initialize() method. 

=back

=cut

######################################################################

sub new { 

   my $proto = shift;
   my $class = ref($proto) || $proto;  # Object or class name

   # Attributes
   my $self = {
       CLASSNAME => "CQCC::TriggerBase",
       OPERATION =>       # Operation, may be overridden by cmdline
           &GetParm("CLEARCASE_OP_KIND"),
       POSTOP => 0,      # Command line flag for post-operations
       ARGSLIST => undef, # Parsed argument list for operation
       ENTITYDEF =>       # Default record type
           &GetParm("CQCC_ENTITY"),
       CACHE_FILE =>      # Name of cache file, set below
           undef, 
       CC => undef,       # ClearCase interface object     
       CQ => undef,       # ClearQuest interface object     
       HELP_CONTEXT => "",# Help topic in case help is invoked 
       CONFIG_PATH =>      # Pathname to configuration script
             undef,
   };

   my @args = @_;
   while (@args) {
      my $key = shift @args;
      $self->{$key} = shift @args;
   } 
   $self->{CACHE_FILE} = &IOPathname("PARM", "cqcc_cache");
   $self->{CACHE_SERIES_FILE} = &IOPathname("PARM_SERIES", ".cqcc_cache");

   bless($self, $class);
   return $self;
}

######################################################################
#::method Op_classtest

=over 6

=item C<Op_classtest()>

This internal method is invoked to ask a specific class to run its
ClassTest method. At least one argument must be provided to specify
which class is to be tested; additional arguments will be passed to
the ClassTest method. The arguments must be delimited by commas with
no spaces in between. Example:

-op classtest CQCC::CCase,vobname,elemname

=back

=cut 

######################################################################

sub Op_classtest {
    my ($self, @args) = @_;

    my $class = shift @args;

    print "\n****************************************\n";
    print "::Classtest: $class\n\n";

    my @quotedArgs;
    for (@args) {
       push @quotedArgs, "\"$_\"";
    } 
    my $argstring = join(",",@quotedArgs);

    my $command = "$class" . "::ClassTest($argstring)";

    my $t1 = &TimerStart("require $class");
    my $package = eval "require $class";
    &TimerStop($t1);

    $t1 = &TimerStart("$class ->ClassTest()");
    my $result = eval $command;
    &TimerStop($t1);

    return $result;
}

######################################################################
#::function PreprocessArguments

=over 6

=item C<PreprocessArguments()>

This function looks through the incoming arguments for any that must
be pre-processed before the trigger is instantiated, i.e. the
TriggerInterface command line argument to override the trigger
subclass.

=back

=cut 

######################################################################

sub PreprocessArguments {
    my @args = @ARGV;
    my $arg;

    # Process parameter settings -p name,value

    while ($arg = shift @args) {
        if ($arg eq "-p") {
            my $pair = shift @args;
            my ($name,$value) = split(',',$pair);
            if ($name && $value) {
                &SetCmdParm($name,$value);
             }
         }
    }
}

######################################################################
#::method PrintState

=over 6

=item C<PrintState(option)>

This method prints out information on classes, parameters, and
other state information. 

=back

=cut 

######################################################################

sub PrintState {
    my ($self,$option) = @_;

    if ($option eq "CLASSES") {
        &IOPrint("Trigger::PrintState\n");
        &IOPrint("Configuration Path: $self->{CONFIG_PATH}\n");
        my @classList = ("AssocList", "CCase", "CCPerlAPI",
            "CQPerlAPI", "CQSchema", "CQuest", "CQWeb", "IOBase",
            "IOClearPrompt", "MsgCat", "MsgCat_English", "Parm",
            "Timer", "TriggerBase", "TriggerCQCC");
        for (@classList) {
            my $className = "CQCC::$_";
            my $version = "(unloaded)";
            my $cmd = "$className::Version()";
            eval {
                $version = eval $cmd;
            };
            &IOPrint("Class: $className $version");
        }
    } elsif ($option eq "PARMS") {
        &PrintParms();
    }
    return;
}

######################################################################
#::method ProcessArguments

=over 6

=item C<ProcessArguments()>

This method processes command line arguments including -help,
-test, and -required as well as -p name,value pairs; if no 
operation is specified the integration will fail and exit.  

=back

=cut 

######################################################################

sub ProcessArguments {
    my $self = shift;
    my @args = @ARGV;

    my $arg;

    while ($arg = shift @args) { 
        if ($arg eq "-required") {
            &ForceParm("REQUIRED","TRUE");

        } elsif ($arg eq "-postop") {
            $self->{POSTOP} = 1;

        } elsif ($arg eq "-silent") {
            &ForceParm("SILENT","TRUE");

        } elsif ($arg eq "-test") {
            &ForceParm("TESTMODE","TRUE");
            $self->{OPERATION} = "test";
            # Make sure DEBUG is at LEAST at level 1 for testing
            if (! &GetParm("CQCC_DEBUG")) {
                &ForceParm("CQCC_DEBUG",1);
            }

        } elsif ($arg eq "-entitydef") {
            $self->{ENTITYDEF} = shift @args;

        } elsif ($arg eq "-help") {
            &IOMessage("INFO", ["TBASE_I_HELP"]);
            &IOExitHandler(0);

        } elsif ($arg eq "-op") {
            $self->{OPERATION} = shift @args;

            my $opArgs = shift @args;

            # If next token starts with "-" then it's not args
            if (substr($opArgs,0,1) eq "-") {
                unshift @args, $opArgs;
            } else {
                # Split arguments when list is retrieved
                $self->{ARGSLIST} = $opArgs;  
            }
            #Test for known operation (?) else error

        # Process parameter settings -p name,value (moved to 
        #    PreprocessArguments in TriggerBase.

        } elsif ($arg eq "-p") {
            my $pair = shift @args;
          
        # Path argument for informational purposes.
        } elsif ($arg eq "-path") {
            $self->{CONFIG_PATH} = shift @args;

        } elsif ($arg eq "-usage") {
            &IOMessage("INFO",
               ["TBASE_I_USAGE", "cqcc_launch", "cqcc_launch", 
            "cqcc_launch" ]);
            &IOExitHandler(0);

        } else {
            &IOMessage("FATAL", ["TBASE_E_UNKNOWN_ARG", $arg]);
        } 
    }
   
    # Test for adequate settings to run

    if (! defined $self->{OPERATION}) {
        &IOMessage("FATAL", ["TBASE_E_NO_OPERATION"]);
    }

    &IODebugPrint("BASIC", 
                  "TriggerBase::ProcessArguments: " . 
                      "Operation: $self->{OPERATION}" . 
                  "($self->{ARGSLIST})" . 
                  ($self->{POSTOP} ? " PostOp" : "") . "\n   " . 
                  (&GetParm("REQUIRED") eq "TRUE" ? " (Required)" : " (Optional)") .
                  (&GetParm("TESTMODE") eq "TRUE" ? " (TestMode)" : "") .
                  ($self->{ENTITYDEF} ? " Entity: $self->{ENTITYDEF}" 
                     : "")
                  );
}

######################################################################
#::method SaveCache

=over 6

=item C<SaveCache()>

This method saves the values of the parameters named in the 
CACHE instance variable to a file named by the CACHE_FILE
instance variable. 

=back

=cut 

######################################################################

sub SaveCache {
    my $self = shift;
    # TODO: Needs better error handling for two files 
    my $result = 1;
    my @parm_strings;

    ##################################################
    # Write Session wide cache if preferences changed
    # or if there is no session cache yet

    my @userChangedParms = &GetParmNameList("USER_CHANGED_PREF");
    my $sessionCache = $self->{CACHE_FILE};
    
    if (! -e $sessionCache || @userChangedParms) {
        my @cache_parms = &GetParmNameList("SRC_USER_PREF");
        for (@cache_parms) { 
            push @parm_strings, &ParmString($_);
        }
        &IODebugPrint("DETAILED", "SaveCache writing $sessionCache\n" . 
                      "   changed: " . join(",",@userChangedParms));
        $result = &IOFileWrite($sessionCache, @parm_strings);
    }

    ##################################################
    # Write Series cache if it is still required

    if (&GetParm("SAVE_SERIES_CACHE") eq "TRUE") {
        my $seriesCache = $self->{CACHE_SERIES_FILE};
        my @cache_parms = &GetParmNameList("SRC_CACHE");
        &ForceParm("CACHE_SERIES", &GetParm("CLEARCASE_SERIES_ID"));

        for (@cache_parms) { 
            push @parm_strings, &ParmString($_);
        }
        &IODebugPrint("DETAILED", "SaveCache writing $seriesCache\n");
        $result = &IOFileWrite($seriesCache, @parm_strings);

        ########################################
        # If No series id for pre-op checkin then 
        # write the name of the series cache to a cache file based on 
        # the ClearCase PID and current filename that can be retrieved later 
        # May be written more often than needed (mainly needed when 
        # PostCI Verify/Commit is enabled) but other conditions can force 
        # post-checkin activity as well so write whenever it might be needed.

        my $ccPid = $ENV{"CLEARCASE_PPID"};
        my $series = $ENV{"CLEARCASE_SERIES_ID"};
        if ((! defined $series || $series eq "") &&
            ($self->{POSTOP} eq 0 && $ENV{"CLEARCASE_OP_KIND"} eq "checkin")) {
            if (defined $ccPid && $ccPid ne "") {
                my $cc_cache = &IOPathname("PARM",".ccase_cache.$ccPid");
                &IOFileWrite($cc_cache, $seriesCache, $ENV{"CLEARCASE_XPN"});
                &IODebugPrint("DETAILED","SaveCache - writing .ccase_cache.$ccPid");
            } else {
                &IOMessage("FATAL", ["TBASE_E_SAVE_CACHE_SERIES"]);
            }
         }
    }
    return $result;
}

######################################################################
#::Tail
######################################################################

1;
